# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.oauth2.errors import ServerError
from ghoauth.oauth2.token import (
    authorization_code_factory,
    bearer_token_factory,
)
from ghoauth.uriutils import extend_uri_with_params

from .abc import AbstractBaseResponseTypeHandler


class MultipleValuedResponseType(AbstractBaseResponseTypeHandler):
    """
    http://openid.net/specs/oauth-v2-multiple-response-types-1_0.html#ResponseModes
    """

    __slots__ = ["code_factory", "token_factory"]

    def __init__(
        self,
        code_factory=authorization_code_factory,
        token_factory=bearer_token_factory,
    ):
        self.code_factory = code_factory
        self.token_factory = token_factory

    def supported_response_types(self):
        return ["code token"]

    def get_default_response_mode(self, request):
        return "fragment"

    def is_response_mode_allowed(self, response_mode, request):
        return response_mode == "fragment"

    def handle_authorization_request(self, request, repository, subject):
        try:
            code = self.code_factory(
                subject, request.o2_scope, request.o2_client_id, request
            )
            if request.o2_state:
                code["state"] = request.o2_state
            repository.save_authorization_code(
                code, request.o2_redirect_uri, request.o2_client_id, request
            )

            token = self.token_factory(
                subject, request.o2_scope, False, request
            )
            repository.save_access_token(token, request)
        except BaseException as why:
            raise ServerError(request) from why
        resp = code.copy()
        resp.update(token.copy())

        redirect_uri = extend_uri_with_params(
            request.o2_redirect_uri, resp, request.o2_response_mode
        )
        request.response.status = 302
        request.response.location = redirect_uri
        return None



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Optional

from webob.response import Response

from ghoauth.uriutils import extend_uri_with_params


class OAuth2Error(Exception):
    pass


class OAuth2ErrorResponse(Response, OAuth2Error):
    o2_error: Optional[str] = None
    o2_code = 400
    o2_error_description: Optional[str] = None

    def __init__(self, request, error_uri=None):
        Response.__init__(self)
        OAuth2Error.__init__(self)
        self.o2_error_uri = error_uri

        self.o2_redirect_uri = request.o2_redirect_uri
        self.o2_response_mode = request.o2_response_mode
        self.o2_prepared = False

        dct = {"error": self.o2_error}
        if self.o2_error_description:
            dct["error_description"] = self.o2_error_description
        if self.o2_error_uri:
            dct["error_uri"] = self.o2_error_uri
        if request.o2_state:
            dct["state"] = request.o2_state
        self.o2_errdct = dct

    def o2_use_redirect(self):
        if self.o2_prepared:
            return
        self.status_code = 302
        self.location = extend_uri_with_params(
            self.o2_redirect_uri, self.o2_errdct, self.o2_response_mode
        )
        self.o2_prepared = True

    def o2_use_json(self):
        if self.o2_prepared:
            return
        self.status_code = self.o2_code
        self.content_type = "application/json"
        body = json.dumps(self.o2_errdct, ensure_ascii=True, indent=2)
        self.app_iter = [body]
        self.text = body
        self.o2_prepared = True

    def o2_ensure_prepared(self):
        self.o2_use_json()

    def __call__(self, environ, start_response):
        self.o2_ensure_prepared()
        return Response.__call__(self, environ, start_response)


class InvalidRequestError(OAuth2ErrorResponse):
    o2_error = "invalid_request"


class UnauthorizedClientError(OAuth2ErrorResponse):
    o2_error = "unauthorized_client"


class AccessDeniedError(OAuth2ErrorResponse):
    o2_error = "access_denied"


class UnsupportedResponseTypeError(OAuth2ErrorResponse):
    o2_error = "unsupported_response_type"


class InvalidScopeError(OAuth2ErrorResponse):
    o2_error = "invalid_scope"


class ServerError(OAuth2ErrorResponse):
    o2_code = 500
    o2_error = "server_error"


class TemporarilyUnavailableError(OAuth2ErrorResponse):
    o2_code = 503
    o2_error = "temporarily_unavailable"


class UnsupportedGrantTypeError(OAuth2ErrorResponse):
    o2_error = "unsupported_grant_type"


class InvalidClientError(OAuth2ErrorResponse):
    o2_error = "invalid_client"


class InvalidGrantError(OAuth2ErrorResponse):
    o2_error = "invalid_grant"


class FatalOAuth2Error(OAuth2Error):
    pass



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AbstractBaseEndpoint:
    __slots__ = ["repository", "validator", "_handlers"]

    def __init__(self, repository, validator):
        self.repository = repository
        self.validator = validator
        self._handlers = {}

    def _get_handler(self, handler_type):
        return self._handlers[handler_type]

    def register_handler(self, handler, handler_types):
        for handler_type in handler_types:
            self._handlers[handler_type] = handler

    def validate_request(self, request):
        raise NotImplementedError("subclasses must implement this")

    def handle_request(self, request, *args, **kwargs):
        raise NotImplementedError("subclasses must implement this")



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.oauth2.request import (
    BaseRequest,
    OAuth2RequestMixin,
    request_parameter,
)


class OpenIDConnectRequestMixin(OAuth2RequestMixin):
    # § 3.1.2.1
    o2_nonce = request_parameter("nonce")
    o2_display = request_parameter("display")
    o2_prompt = request_parameter("prompt")
    o2_max_age = request_parameter("max_age")
    o2_ui_locales = request_parameter("ui_locales")
    o2_id_token_hint = request_parameter("id_token_hint")
    o2_login_hint = request_parameter("login_hint")
    o2_acr_values = request_parameter("acr_values")
    # § 5.2
    o2_claims_locales = request_parameter("claims_locales")
    # § 5.5
    o2_claims = request_parameter("claims")
    # § 6
    o2_request = request_parameter("request")
    o2_request_uri = request_parameter("request_uri")


OpenIDConnectRequest = type(
    "OpenIDConnectRequest", (OpenIDConnectRequestMixin, BaseRequest), {}
)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.oauth2.errors import (
    InvalidGrantError,
    InvalidRequestError,
    InvalidScopeError,
    ServerError,
)
from ghoauth.oauth2.token import bearer_token_factory

from .abc import AbstractBaseGrantTypeHandler


class RefreshTokenGrant(AbstractBaseGrantTypeHandler):
    """
    https://tools.ietf.org/html/rfc6749#section-6
    """

    __slots__ = ["issue_refresh_token", "token_factory"]

    def __init__(
        self, issue_refresh_token=True, token_factory=bearer_token_factory
    ):
        self.issue_refresh_token = issue_refresh_token
        self.token_factory = token_factory

    def supported_grant_types(self):
        return ["refresh_token"]

    def validate_token_request(self, request, repository, validator):
        if not request.o2_validate(
            requireds={"refresh_token"}, optionals={"scope"}
        ):
            raise InvalidRequestError(request)
        if not validator.validate_refresh_token(
            request.o2_refresh_token, request.o2_client_id, request
        ):
            raise InvalidGrantError(request)

        granted_scope = repository.get_scope_by_refresh_token(
            request.o2_refresh_token, request
        )
        if request.o2_scope:
            if not validator.validate_scope(request.o2_scope, request):
                raise InvalidScopeError(request)
            if not validator.is_scope_subset_of(
                request.o2_scope, granted_scope
            ):
                raise InvalidScopeError(request)
        else:
            request.o2_scope = granted_scope

    def handle_token_request(self, request, repository):
        try:
            subject = repository.get_subject_by_refresh_token(
                request.o2_refresh_token
            )
            token = self.token_factory(
                subject, request.o2_scope, self.issue_refresh_token, request
            )
            repository.save_access_token(token, request)
        except BaseException as why:
            raise ServerError(request) from why

        return token



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import UserDict


class OpenIDConnectProviderMetadata(UserDict):
    _attributes = [
        ("issuer", None, True),
        ("authorization_endpoint", None, True),
        ("token_endpoint", None, False),
        ("userinfo_endpoint", None, False),
        ("jwks_uri", None, True),
        ("registration_endpoint", None, False),
        ("scopes_supported", None, False),
        ("response_types_supported", None, True),
        ("response_modes_supported", ["query", "fragment"], False),
        ("grant_types_supported", ["authorization_code", "implicit"], False),
        ("acr_values_supported", None, False),
        ("subject_types_supported", None, True),
        ("id_token_signing_alg_values_supported", None, True),
        ("id_token_encryption_alg_values_supported", None, False),
        ("id_token_encryption_enc_values_supported", None, False),
        ("userinfo_signing_alg_values_supported", None, False),
        ("userinfo_encryption_alg_values_supported", None, False),
        ("userinfo_encryption_enc_values_supported", None, False),
        ("request_object_signing_alg_values_supported", None, False),
        ("request_object_encryption_alg_values_supported", None, False),
        ("request_object_encryption_enc_values_supported", None, False),
        (
            "token_endpoint_auth_methods_supported",
            ["client_secret_basic"],
            False,
        ),
        ("token_endpoint_auth_signing_alg_values_supported", None, False),
        ("display_values_supported", None, False),
        ("claim_types_supported", ["normal"], False),
        ("claims_supported", None, False),
        ("service_documentation", None, False),
        ("claims_locales_supported", None, False),
        ("ui_locales_supported", None, False),
        ("claims_parameter_supported", False, False),
        ("request_parameter_supported", False, False),
        ("request_uri_parameter_supported", True, False),
        ("require_request_uri_registration", False, False),
        ("op_policy_uri", None, False),
        ("op_tos_uri", None, False),
    ]

    def __init__(self, issuer):
        UserDict.__init__(self)
        self.data["issuer"] = issuer

    def set_authorization_endpoint(self, value):
        self.data["authorization_endpoint"] = value

    def set_token_endpoint(self, value):
        self.data["token_endpoint"] = value

    def set_userinfo_endpoint(self, value):
        self.data["userinfo_endpoint"] = value

    def set_jwks_uri(self, value):
        self.data["jwks_uri"] = value

    def set_registration_endpoint(self, value):
        self.data["registration_endpoint"] = value

    def set_scopes_supported(self, value):
        self.data["scopes_supported"] = value

    def set_response_types_supported(self, value):
        self.data["response_types_supported"] = value

    def set_response_modes_supported(self, value):
        self.data["response_modes_supported"] = value

    def set_grant_types_supported(self, value):
        self.data["grant_types_supported"] = value

    def set_acr_values_supported(self, value):
        self.data["acr_values_supported"] = value

    def set_subject_types_supported(self, value):
        self.data["subject_types_supported"] = value

    def set_id_token_signing_alg_values_supported(self, value):
        self.data["id_token_signing_alg_values_supported"] = value

    def set_id_token_encryption_alg_values_supported(self, value):
        self.data["id_token_encryption_alg_values_supported"] = value

    def set_id_token_encryption_enc_values_supported(self, value):
        self.data["id_token_encryption_enc_values_supported"] = value

    def set_userinfo_signing_alg_values_supported(self, value):
        self.data["userinfo_signing_alg_values_supported"] = value

    def set_userinfo_encryption_alg_values_supported(self, value):
        self.data["userinfo_encryption_alg_values_supported"] = value

    def set_userinfo_encryption_enc_values_supported(self, value):
        self.data["userinfo_encryption_enc_values_supported"] = value

    def set_request_object_signing_alg_values_supported(self, value):
        self.data["request_object_signing_alg_values_supported"] = value

    def set_request_object_encryption_alg_values_supported(self, value):
        self.data["request_object_encryption_alg_values_supported"] = value

    def set_request_object_encryption_enc_values_supported(self, value):
        self.data["request_object_encryption_enc_values_supported"] = value

    def set_token_endpoint_auth_methods_supported(self, value):
        self.data["token_endpoint_auth_methods_supported"] = value

    def set_token_endpoint_auth_signing_alg_values_supported(self, value):
        self.data["token_endpoint_auth_signing_alg_values_supported"] = value

    def set_display_values_supported(self, value):
        self.data["display_values_supported"] = value

    def set_claim_types_supported(self, value):
        self.data["claim_types_supported"] = value

    def set_claims_supported(self, value):
        self.data["claims_supported"] = value

    def set_service_documentation(self, value):
        self.data["service_documentation"] = value

    def set_claims_locales_supported(self, value):
        self.data["claims_locales_supported"] = value

    def set_ui_locales_supported(self, value):
        self.data["ui_locales_supported"] = value

    def set_claims_parameter_supported(self, value):
        self.data["claims_parameter_supported"] = value

    def set_request_parameter_supported(self, value):
        self.data["request_parameter_supported"] = value

    def set_request_uri_parameter_supported(self, value):
        self.data["request_uri_parameter_supported"] = value

    def set_require_request_uri_registration(self, value):
        self.data["require_request_uri_registration"] = value

    def set_op_policy_uri(self, value):
        self.data["op_policy_uri"] = value

    def set_op_tos_uri(self, value):
        self.data["op_tos_uri"] = value

    def validate(self):
        for name, _default, required in self._attributes:
            if not required:
                continue
            if name not in self.data:
                raise ValueError("{} is required".format(name))

        # token_endpoint is REQUIRED unless only the Implicit Flow is used.
        grant_types = self.data.get("grant_types_supported", [])
        if (
            len(grant_types) != 1 or grant_types[0] != "implicit"
        ) and "token_endpoint" not in self.data:
            raise ValueError("token_endpoint is required")

    def to_dict(self, with_defaults=False):
        ret = {}
        for name, default, required in self._attributes:
            if name in self.data:
                ret[name] = self.data[name]
                continue
            if with_defaults and default is not None:
                ret[name] = default

        additionals = set(self.data.keys()) - set(ret.keys())
        for name in additionals:
            ret[name] = self.data[name]

        return ret



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from base64 import b64encode


def generate_token():
    """
    The probability of an attacker guessing generated tokens (and other
    credentials not intended for handling by end-users) MUST be less than
    or equal to 2^(-128) and SHOULD be less than or equal to 2^(-160).
    https://tools.ietf.org/html/rfc6749#section-10.10
    """
    return b64encode(os.urandom(20)).decode("ascii")


def authorization_code_factory(subject, scope, client_id, request):
    """
    The authorization code MUST expire shortly after it is issued
    to mitigate the risk of leaks.  A maximum authorization code
    lifetime of 10 minutes is RECOMMENDED
    https://tools.ietf.org/html/rfc6749#section-4.1.2
    """
    return {"code": generate_token()}


def bearer_token_factory(subject, scope, with_refresh_token, request):
    dct = {
        "token_type": "Bearer",
        "scope": scope,
        "access_token": generate_token(),
        "expires_in": 3600,
    }
    if with_refresh_token:
        dct["refresh_token"] = generate_token()
    return dct



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from jose import (
    JWTError,
    jwt,
)

from .grant_types.jwt_bearer import decode_jwt_assertion


class JWTClientAuthenticator:

    __slots__ = ["repository", "validator", "datetime_factory"]

    def __init__(self, repository, validator, datetime_factory=datetime.now):
        self.repository = repository
        self.validator = validator
        self.datetime_factory = datetime_factory

    def __call__(self, request):
        if not request.o2_validate(
            requireds={"client_assertion_type", "client_assertion_type"}
        ):
            return False
        if (
            request.o2_client_assertion_type
            != "urn:ietf:params:oauth:client-assertion-type:jwt-bearer"
        ):
            return False

        try:
            headers = jwt.get_unverified_headers(request.o2_assertion)
            jwt_claims = jwt.get_unverified_claims(request.o2_client_assertion)
            client_id = jwt_claims["sub"]
            if not self.repository.is_client_available(client_id, request):
                return False
            decode_key = self.repository.get_key_by_jwt_headers(
                headers, client_id, request
            )
        except JWTError:
            return False
        else:
            jwt_claims = decode_jwt_assertion(
                request.o2_client_assertion,
                decode_key,
                request,
                self.validator,
                self.datetime_factory().timestamp(),
                False,
            )
            if not jwt_claims:
                return False

        return client_id



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import (
    parse_qsl,
    urlencode,
)

from rfc3986 import (
    URIReference,
    uri_reference,
    urlparse,
)

__all__ = [
    "compare_uri",
    "extend_uri_with_params",
    "is_absolute_uri",
    "uri_reference",
    "urlparse",
    "URIReference",
]


def is_absolute_uri(uri):
    if isinstance(uri, URIReference):
        parsed = uri
    else:
        try:
            parsed = uri_reference(uri)
        except BaseException:
            return False

    return parsed.is_absolute() and parsed.scheme and parsed.host


def extend_uri_with_params(redirect_uri, paramdict, response_mode):
    parsed = urlparse(redirect_uri)
    if response_mode == "fragment":
        fragment = urlencode(list(paramdict.items()))
        parsed = parsed.copy_with(fragment=fragment)
    else:
        qsl = parse_qsl(parsed.query, keep_blank_values=True)
        qsl.extend(paramdict.items())
        query = urlencode(qsl)
        parsed = parsed.copy_with(query=query)
    return parsed.geturl()


def compare_uri(x, y):
    try:
        if isinstance(x, URIReference):
            x_uri = x
        else:
            x_uri = uri_reference(x)
        if isinstance(y, URIReference):
            y_uri = y
        else:
            y_uri = uri_reference(y)
    except BaseException:
        return False
    else:
        return x_uri.normalized_equality(y_uri)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from webob import Response

from ghoauth.oauth2 import OAuth2Request
from ghoauth.oauth2.errors import (
    AccessDeniedError,
    ServerError,
)


class OAuth2ErrorTest(TestCase):

    def test_o2_use_redirect(self):
        request = OAuth2Request.blank(
            "https://example.com/authorization"
            "?state=csrftoken"
            "&redirect_uri=https%3A//rp.example.com/callback"
        )
        inst = AccessDeniedError(request)
        inst.o2_use_redirect()

        self.assertTrue(inst.o2_prepared)
        self.assertIsInstance(inst, Response)
        self.assertEqual(inst.status_code, 302)
        self.assertEqual(
            inst.location,
            "https://rp.example.com/callback"
            "?error=access_denied&state=csrftoken",
        )

    def test_o2_use_redirect_fragment(self):
        request = OAuth2Request.blank(
            "https://example.com/authorization"
            "?state=csrftoken"
            "&redirect_uri=https%3A//rp.example.com/callback"
            "&response_mode=fragment"
        )
        inst = AccessDeniedError(request)
        inst.o2_use_redirect()

        self.assertTrue(inst.o2_prepared)
        self.assertIsInstance(inst, Response)
        self.assertEqual(inst.status_code, 302)
        self.assertEqual(
            inst.location,
            "https://rp.example.com/callback"
            "#error=access_denied&state=csrftoken",
        )

    def test_o2_use_json(self):
        request = OAuth2Request.blank(
            "https://example.com/authorization"
            "?state=csrftoken"
            "&redirect_uri=https%3A//rp.example.com/callback"
        )
        inst = ServerError(request, "https://example.com/docs/")
        inst.o2_use_json()

        self.assertTrue(inst.o2_prepared)
        self.assertIsInstance(inst, Response)
        self.assertEqual(inst.status_code, 500)
        self.assertEqual(inst.content_type, "application/json")
        self.assertEqual(
            inst.json,
            {
                "error": "server_error",
                "error_uri": "https://example.com/docs/",
                "state": "csrftoken",
            },
        )



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .client_authenticator import JWTClientAuthenticator
from .grant_types import (
    AuthorizationCodeGrant,
    ImplicitGrant,
    JWTBearerGrant,
    MultipleValuedResponseType,
    RefreshTokenGrant,
)
from .repository import (
    AbstractBaseRepository,
    AuthorizationCodeGrantRepositoryMixin,
    JWTBearerGrantRepositoryMixin,
    RefreshTokenGrantRepositoryMixin,
)
from .request import (
    OAuth2Request,
    OAuth2RequestMixin,
)
from .server import AuthorizationServer
from .token import (
    authorization_code_factory,
    bearer_token_factory,
)
from .validator import (
    AbstractBaseValidator,
    AuthorizationCodeGrantValidatorMixin,
    JWTBearerGrantValidatorMixin,
    RefreshTokenGrantValidatorMixin,
)

__all__ = [
    "AbstractBaseRepository",
    "AbstractBaseValidator",
    "AuthorizationServer",
    "OAuth2Request",
    "OAuth2RequestMixin",
    "authorization_code_factory",
    "bearer_token_factory",
    "AuthorizationCodeGrant",
    "AuthorizationCodeGrantRepositoryMixin",
    "AuthorizationCodeGrantValidatorMixin",
    "ImplicitGrant",
    "JWTBearerGrant",
    "JWTBearerGrantRepositoryMixin",
    "JWTBearerGrantValidatorMixin",
    "MultipleValuedResponseType",
    "RefreshTokenGrant",
    "RefreshTokenGrantRepositoryMixin",
    "RefreshTokenGrantValidatorMixin",
    "JWTClientAuthenticator",
]



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import (
    datetime,
    timezone,
)
from unittest import TestCase

from ghoauth.oauth2.errors import (
    InvalidGrantError,
    InvalidRequestError,
    InvalidScopeError,
)
from ghoauth.oauth2.grant_types.jwt_bearer import JWTBearerGrant
from ghoauth.oauth2.repository import (
    AbstractBaseRepository,
    JWTBearerGrantRepositoryMixin,
)
from ghoauth.oauth2.request import OAuth2Request
from ghoauth.oauth2.validator import (
    AbstractBaseValidator,
    JWTBearerGrantValidatorMixin,
)
from ghoauth.uriutils import compare_uri


class Repository(AbstractBaseRepository, JWTBearerGrantRepositoryMixin):

    def get_default_scope(self, request):
        return "email"

    def get_key_by_jwt_headers(self, headers, client_id, request):
        return "secretstring"

    def get_subject_by_jwt_claims(self, claims, request):
        return "yosida95"

    def save_access_token(self, token, request):
        assert token["scope"] == "email"


class Validator(AbstractBaseValidator, JWTBearerGrantValidatorMixin):

    def validate_scope(self, scope, request):
        return scope == "email"

    def validate_iss(self, iss, request):
        return compare_uri(iss, "https://jwt-idp.example.com")

    def validate_sub(self, sub, request):
        return True

    def validate_aud(self, aud, request):
        return compare_uri(aud, "https://jwt-rp.example.net")


class JWTBearerGrantTeset(TestCase):

    def setUp(self):
        self.repository = Repository()
        self.validator = Validator()
        self.current_dt = datetime(2011, 3, 22, 17, 43, tzinfo=timezone.utc)

    def test_supported_grant_types(self):
        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        self.assertEqual(
            inst.supported_grant_types(),
            ["urn:ietf:params:oauth:grant-type:jwt-bearer"],
        )

    def test_validate_token_request(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJtYWlsdG86bWlrZUBleGFtcGxlLmNvbSIsImV4cCI6MTMwMDgxOTM4MCwibmJmIjoxMzAwODE1NzgwLCJhdWQiOiJodHRwczovL2p3dC1ycC5leGFtcGxlLm5ldCIsImlzcyI6Imh0dHBzOi8vand0LWlkcC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.fTxrgzdzh-sgE1in4OFbGMGTlHMd9idBLL4ezkyBcGw",  # noqa: E501
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        inst.validate_token_request(request, self.repository, self.validator)
        self.assertEqual(
            request._o2_jwt_claims,
            {
                "iss": "https://jwt-idp.example.com",
                "sub": "mailto:mike@example.com",
                "aud": "https://jwt-rp.example.net",
                "nbf": 1300815780,
                "exp": 1300819380,
                "http://claims.example.com/member": True,
            },
        )

        before = datetime(2011, 3, 22, 17, 42, tzinfo=timezone.utc)
        inst = JWTBearerGrant(datetime_factory=lambda: before)
        with self.assertRaises(InvalidGrantError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

        expired = datetime(2011, 3, 22, 18, 44, tzinfo=timezone.utc)
        inst = JWTBearerGrant(datetime_factory=lambda: expired)
        with self.assertRaises(InvalidGrantError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_validate_token_request_missing_assertion(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={"grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer"},
        )
        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        with self.assertRaises(InvalidRequestError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_validate_token_request_with_scope(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJtYWlsdG86bWlrZUBleGFtcGxlLmNvbSIsImV4cCI6MTMwMDgxOTM4MCwibmJmIjoxMzAwODE1NzgwLCJhdWQiOiJodHRwczovL2p3dC1ycC5leGFtcGxlLm5ldCIsImlzcyI6Imh0dHBzOi8vand0LWlkcC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.fTxrgzdzh-sgE1in4OFbGMGTlHMd9idBLL4ezkyBcGw",  # noqa: E501
                "scope": "email",
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        inst.validate_token_request(request, self.repository, self.validator)

    def test_validate_token_request_with_invalid_scope(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJtYWlsdG86bWlrZUBleGFtcGxlLmNvbSIsImV4cCI6MTMwMDgxOTM4MCwibmJmIjoxMzAwODE1NzgwLCJhdWQiOiJodHRwczovL2p3dC1ycC5leGFtcGxlLm5ldCIsImlzcyI6Imh0dHBzOi8vand0LWlkcC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.fTxrgzdzh-sgE1in4OFbGMGTlHMd9idBLL4ezkyBcGw",  # noqa: E501
                "scope": "phone",
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        with self.assertRaises(InvalidScopeError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_validate_token_request_iss_mismatch(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwczovL21hbGljaW91cy1pZHAuZXhhbXBsZS5jb20iLCJuYmYiOjEzMDA4MTU3ODAsInN1YiI6Im1haWx0bzptaWtlQGV4YW1wbGUuY29tIiwiZXhwIjoxMzAwODE5MzgwLCJhdWQiOiJodHRwczovL2p3dC1ycC5leGFtcGxlLm5ldCIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.EY9ygpJP6kbmakVED226wY6bbaNsKyxZg5lrD7vkaYQ",  # noqa: E501
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        with self.assertRaises(InvalidGrantError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_validate_token_request_aud_mismatch(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwczovL2p3dC1pZHAuZXhhbXBsZS5jb20iLCJuYmYiOjEzMDA4MTU3ODAsInN1YiI6Im1haWx0bzptaWtlQGV4YW1wbGUuY29tIiwiZXhwIjoxMzAwODE5MzgwLCJhdWQiOiJodHRwczovL21hbGljaW91cy1ycC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.9ioNkxKlnJapTLBRL9zhiyRdIDOY-TzoA0pSLrF0YXY",  # noqa: E501
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        with self.assertRaises(InvalidGrantError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_validate_token_secret_mismatch(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJtYWlsdG86bWlrZUBleGFtcGxlLmNvbSIsImlzcyI6Imh0dHBzOi8vand0LWlkcC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlLCJleHAiOjEzMDA4MTkzODAsImF1ZCI6Imh0dHBzOi8vand0LXJwLmV4YW1wbGUubmV0IiwibmJmIjoxMzAwODE1NzgwfQ.mtAeY_YoHe-JNK_pzo_KQDE4TAuNTI0TSoKTzQJCp0Q",  # noqa: E501
            },
        )

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        with self.assertRaises(InvalidGrantError):
            inst.validate_token_request(
                request, self.repository, self.validator
            )

    def test_handle_token_request(self):
        request = OAuth2Request.blank(
            "https://example.com/oauth2/token",
            POST={
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJtYWlsdG86bWlrZUBleGFtcGxlLmNvbSIsImV4cCI6MTMwMDgxOTM4MCwibmJmIjoxMzAwODE1NzgwLCJhdWQiOiJodHRwczovL2p3dC1ycC5leGFtcGxlLm5ldCIsImlzcyI6Imh0dHBzOi8vand0LWlkcC5leGFtcGxlLmNvbSIsImh0dHA6Ly9jbGFpbXMuZXhhbXBsZS5jb20vbWVtYmVyIjp0cnVlfQ.fTxrgzdzh-sgE1in4OFbGMGTlHMd9idBLL4ezkyBcGw",  # noqa: E501
            },
        )
        request.response = response = OAuth2Request.ResponseClass()

        inst = JWTBearerGrant(datetime_factory=lambda: self.current_dt)
        inst.validate_token_request(request, self.repository, self.validator)
        inst.handle_token_request(request, self.repository)

        self.assertEqual(response.status, "200 OK")



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.uriutils import (
    is_absolute_uri,
    urlparse,
)


def normalize_response_type(response_type):
    """
    Extension response types MAY contain a space-delimited (%x20) list of
    values, where the order of values does not matter (e.g., response
    type "a b" is the same as "b a").  The meaning of such composite
    response types is defined by their respective specifications.

    https://tools.ietf.org/html/rfc6749#section-3.1.1
    """
    return " ".join(sorted(response_type.split()))


def validate_redirect_uri(redirect_uri):
    """
    The redirection endpoint URI MUST be an absolute URI as defined by
    [RFC3986] Section 4.3.  The endpoint URI MAY include an
    "application/x-www-form-urlencoded" formatted (per Appendix B) query
    component ([RFC3986] Section 3.4), which MUST be retained when adding
    additional query parameters.  The endpoint URI MUST NOT include a
    fragment component.

    https://tools.ietf.org/html/rfc6749#section-3.1.2
    """
    parsed = urlparse(redirect_uri)
    if not is_absolute_uri(parsed.geturl()):
        return False
    if parsed.fragment:
        return False
    return True


class AbstractBaseValidator:

    def authenticate_client(self, request):
        raise NotImplementedError("must implement authenticate_client")

    def authenticate_client_as(self, client_id, request):
        raise NotImplementedError("must implement authenticate_client_as")

    def authorize_response_type(self, response_type, client_id, request):
        raise NotImplementedError("must implement authorize_response_type")

    def authorize_grant_type(self, grant_type, client_id, request):
        raise NotImplementedError("must implement authorize_grant_type")

    def validate_redirect_uri(
        self, redirect_uri, client_id, request, exactly=False
    ):
        """
        The authorization server MUST require the following clients to
        register their redirection endpoint:
            o  Public clients.
            o  Confidential clients utilizing the implicit grant type.
        The authorization server SHOULD require all clients to register their
        redirection endpoint prior to utilizing the authorization endpoint.

        The authorization server SHOULD require the client to provide the
        complete redirection URI (the client MAY use the "state" request
        parameter to achieve per-request customization).  If requiring the
        registration of the complete redirection URI is not possible, the
        authorization server SHOULD require the registration of the URI
        scheme, authority, and path (allowing the client to dynamically vary
        only the query component of the redirection URI when requesting
        authorization).
        The authorization server MAY allow the client to register multiple
        redirection endpoints.

        https://tools.ietf.org/html/rfc6749#section-3.1.2.2
        """
        raise NotImplementedError("must implement validate_redirect_uri")

    def validate_scope(self, scope, client_id, request):
        raise NotImplementedError("must implement validate_scope")


class AuthorizationCodeGrantValidatorMixin:

    def validate_authorization_code(
        self, code, redirect_uri, client_id, request
    ):
        """
        REQUIRED, if the "redirect_uri" parameter was included in the
        authorization request as described in Section 4.1.1, and their
        values MUST be identical.

        https://tools.ietf.org/html/rfc6749#section-4.1.3
        """
        raise NotImplementedError("must implement validate_authorization_code")


class RefreshTokenGrantValidatorMixin:

    def validate_refresh_token(self, refresh_token, clieht_id, request):
        raise NotImplementedError("must implement validate_refresh_token")

    def is_scope_subset_of(self, subset, superset):
        """
        The requested scope MUST NOT include any scope
        not originally granted by the resource owner

        https://tools.ietf.org/html/rfc6749#section-6
        """
        raise NotImplementedError("must implement is_scope_subset_of")


class JWTBearerGrantValidatorMixin:

    def validate_iss(self, iss, request):
        """
        The JWT MUST contain an "iss" (issuer) claim that contains a
        unique identifier for the entity that issued the JWT.  In the
        absence of an application profile specifying otherwise,
        compliant applications MUST compare issuer values using the
        Simple String Comparison method defined in Section 6.2.1 of RFC
        3986
        """
        raise NotImplementedError("must implement validate_iss")

    def validate_sub(self, sub, request, is_authorization_grant):
        """
        The JWT MUST contain a "sub" (subject) claim identifying the
        principal that is the subject of the JWT.  The subject typically
        identifies an authorized accessor for which the access token
        is being requested (i.e., the resource owner or an
        authorized delegate), but in some cases, may be a
        pseudonymous identifier or other value denoting an anonymous
        user.
        """
        raise NotImplementedError("must implement validate_sub")

    def validate_aud(self, aud, request):
        """
        The JWT MUST contain an "aud" (audience) claim containing a
        value that identifies the authorization server as an intended
        audience.
        The authorization server MUST reject any JWT that does not contain
        its own identity as the intended audience.  In the absence of an
        application profile specifying otherwise, compliant applications
        MUST compare the audience values using the Simple String
        Comparison method defined in Section 6.2.1 of RFC 3986.
        """
        raise NotImplementedError("must implement validate_aud")

    def validate_jti(self, jti, request):
        """
        The JWT MAY contain a "jti" (JWT ID) claim that provides a
        unique identifier for the token.  The authorization server MAY
        ensure that JWTs are not replayed by maintaining the set of used
        "jti" values for the length of time for which the JWT would be
        considered valid based on the applicable "exp" instant.
        """



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from ghoauth.oauth2.validator import (
    normalize_response_type,
    validate_redirect_uri,
)


class normalize_response_type_test(TestCase):

    def test_it(self):
        self.assertEqual(normalize_response_type("code"), "code")
        self.assertEqual(normalize_response_type("code token"), "code token")
        self.assertEqual(normalize_response_type("token code"), "code token")


class validate_redirect_uri_test(TestCase):

    def test_it(self):
        self.assertTrue(validate_redirect_uri("https://example.com/cb"))
        self.assertFalse(
            validate_redirect_uri("urn:example:animal:ferret:nose")
        )

        # with fragment
        self.assertFalse(validate_redirect_uri("https://example.com/cb#k=v"))
        # relative uri
        self.assertFalse(validate_redirect_uri("/cb"))



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from webob import BaseRequest

from ghoauth.oauth2.repository import AbstractBaseRepository
from ghoauth.oauth2.request import (
    OAuth2Request,
    request_parameter,
)
from ghoauth.oauth2.validator import AbstractBaseValidator


class RequestParameterTest(TestCase):

    class Request(BaseRequest):
        o2_param = request_parameter("param")
        o2_dup = request_parameter("dup")

    def test_it(self):
        req = self.Request.blank(
            "https://example.com/?param=value&dup=1&dup=2"
        )
        self.assertNotIn("oauth2.params", req.environ)

        self.assertEqual(req.o2_param, "value")
        self.assertIsNone(req.o2_dup)

        self.assertIn("oauth2.params", req.environ)
        self.assertEqual(
            req.environ["oauth2.params"], {"param": "value", "dup": None}
        )

        req.o2_param = "new_value"
        self.assertEqual(req.o2_param, "new_value")
        self.assertIsNone(req.o2_dup)
        self.assertEqual(
            req.environ["oauth2.params"], {"param": "new_value", "dup": None}
        )

        with self.assertRaises(AttributeError):
            del req.o2_dup


class OAuth2RequestTest(TestCase):

    class StubRepository(AbstractBaseRepository):

        def is_client_available(self, client_id, request):
            return client_id == "yosida95"

        def get_primary_redirect_uri(self, client_id, request):
            if client_id == "yosida95":
                return "https://yosida95.com/cb"
            return "/cb"

    class StubValidator(AbstractBaseValidator):

        def authenticate_client(self, request):
            return "yosida95"

        def authenticate_client_as(self, client_id, request):
            return client_id == "yosida95"

        def validate_redirect_uri(
            self, redirect_uri, client_id, request, exactly=False
        ):
            return redirect_uri == "https://yosida95.com/cb"

    def test_o2_validate(self):
        req = OAuth2Request.blank(
            "https://example.com/?client_id=yosida95&response_type=code&scope=abc&scope=def"  # noqa: E501
        )

        self.assertEqual(req.o2_client_id, "yosida95")
        self.assertEqual(req.o2_response_type, "code")

        # requireds
        self.assertTrue(
            req.o2_validate(requireds={"client_id", "response_type"})
        )
        # optionals
        self.assertTrue(req.o2_validate(optionals={"redirect_uri"}))
        # mix
        self.assertTrue(
            req.o2_validate(
                requireds={"client_id", "response_type"},
                optionals={"redirect_uri"},
            )
        )

        # a missing parameter
        self.assertFalse(req.o2_validate(requireds={"redirect_uri"}))
        # a duplicated parameter
        self.assertFalse(req.o2_validate(requireds={"scope"}))
        self.assertFalse(req.o2_validate(optionals={"scope"}))

    def test_o2_validate_redirect_uri(self):
        repo = self.StubRepository()
        valid = self.StubValidator()

        # use valid primary
        req = OAuth2Request.blank("https://example.com/?client_id=yosida95")
        self.assertTrue(req.o2_validate_redirect_uri(repo, valid))
        self.assertEqual(req.o2_redirect_uri, "https://yosida95.com/cb")

        # use invalid primary
        req = OAuth2Request.blank("https://example.com/?client_id=yoshida95")
        self.assertFalse(req.o2_validate_redirect_uri(repo, valid))
        self.assertEqual(req.o2_redirect_uri, "/cb")

        # use query parameters with a valid value
        req = OAuth2Request.blank(
            "https://example.com/?client_id=yosida95&redirect_uri=https://yosida95.com/cb"  # noqa: E501
        )
        self.assertTrue(req.o2_validate_redirect_uri(repo, valid))
        self.assertEqual(req.o2_redirect_uri, "https://yosida95.com/cb")

        # use query parameters with a invalid value
        req = OAuth2Request.blank(
            "https://example.com/?client_id=yosida95&redirect_uri=/cb"
        )
        self.assertFalse(req.o2_validate_redirect_uri(repo, valid))
        self.assertEqual(req.o2_redirect_uri, "/cb")



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from typing import (
    Any,
    Callable,
    Dict,
    List,
)

from ghoauth.oauth2.errors import (
    FatalOAuth2Error,
    InvalidGrantError,
    InvalidRequestError,
    InvalidScopeError,
    ServerError,
)
from ghoauth.oauth2.grant_types.abc import (
    AbstractBaseGrantTypeHandler,
    AbstractBaseResponseTypeHandler,
)
from ghoauth.oauth2.request import OAuth2RequestMixin
from ghoauth.oauth2.token import (
    authorization_code_factory,
    bearer_token_factory,
)
from ghoauth.uriutils import extend_uri_with_params
from jwt import JWT
from jwt.utils import b64encode

from .repository import AbstractOpenIDRepository
from .request import OpenIDConnectRequestMixin
from .validator import AbstractOpenIDRequestValidator


def _std_hash_by_alg(alg) -> Callable[[bytes], Any]:
    if alg.endswith("S256"):
        return hashlib.sha256
    if alg.endswith("S384"):
        return hashlib.sha384
    if alg.endswith("S512"):
        return hashlib.sha512
    raise ValueError("{} is not supported".format(alg))


def _left_half_hash(data: str, alg: str) -> str:
    hashfun = _std_hash_by_alg(alg)
    hashobj = hashfun(data.encode("ascii"))
    digest = hashobj.digest()[: hashobj.digest_size // 2]
    return b64encode(digest)


class OpenIDConnectGrant(
    AbstractBaseResponseTypeHandler, AbstractBaseGrantTypeHandler
):
    """
    http://openid.net/specs/openid-connect-core-1_0.html
    """

    __slots__ = ["code_factory", "token_factory", "jwt_encoder"]
    jwt_encoder_cls = JWT

    def __init__(
        self,
        code_factory: Callable[
            [Any, str, str, OAuth2RequestMixin], Dict[str, Any]
        ] = authorization_code_factory,
        token_factory: Callable[
            [Any, str, bool, OAuth2RequestMixin], Dict[str, Any]
        ] = bearer_token_factory,
    ) -> None:
        self.code_factory = code_factory
        self.token_factory = token_factory
        self.jwt_encoder = self.jwt_encoder_cls()

    def supported_response_types(self) -> List[str]:
        return [
            # Authorization Code Flow
            "code",
            # Implicit Flow
            "id_token",
            "id_token token",
            # Hybrid Flow
            "code id_token",
            "code token",
            "code id_token token",
            # Multiple Response Type Encoding Practices
            "none",
            # Oops
            "token",
        ]

    def get_default_response_mode(
        self, request: OpenIDConnectRequestMixin
    ) -> str:
        if request.o2_response_type in ("code", "none"):
            return "query"
        return "fragment"

    def is_response_mode_allowed(
        self, response_mode: str, request: OpenIDConnectRequestMixin
    ) -> bool:
        """
        http://openid.net/specs/oauth-v2-multiple-response-types-1_0.html#Combinations
        """
        if request.o2_response_type in ("code", "none"):
            return response_mode in ("query", "fragment")
        return response_mode == "fragment"

    def validate_authorization_request(
        self,
        request: OpenIDConnectRequestMixin,
        repository: AbstractOpenIDRepository,
        validator: AbstractOpenIDRequestValidator,
    ) -> None:
        if (
            not request.o2_validate(requireds={"scope"})
            or "openid" not in request.o2_scope.split()
        ):
            # the request is not for Open ID Connect.
            AbstractBaseResponseTypeHandler.validate_authorization_request(
                self, request, repository, validator
            )
            return

        if not request.o2_validate(requireds={"client_id", "redirect_uri"}):
            raise FatalOAuth2Error()
        if not repository.is_client_available(request.o2_client_id, request):
            raise FatalOAuth2Error()
        # NOTE(yosida95): 'response_type' parameter was eventually validated by
        # the AuthorizationEndpoint to dispatch this method, the value would be
        # one of the self.supported_response_types(), so the value can be used
        # safely here.
        resp_type = request.o2_response_type.split()
        if not request.o2_validate_redirect_uri(
            repository, validator, exactly="code" not in resp_type
        ):
            raise FatalOAuth2Error()

        if not request.o2_validate(
            optionals={
                "state",
                "nonce",
                "display",
                "prompt",
                "max_age",
                "ui_locales",
                "id_token_hint",
                "login_hint",
                "acr_values",
            }
        ):
            raise InvalidRequestError(request)
        if "code" not in resp_type and not request.o2_nonce:
            # In implicit flow, use of nonce is REQUIRED.
            raise InvalidRequestError(request)

        if not validator.validate_scope(
            request.o2_scope, request.o2_client_id, request
        ):
            raise InvalidScopeError(request)

    def handle_authorization_request(
        self,
        request: OpenIDConnectRequestMixin,
        repository: AbstractOpenIDRepository,
        subject: Any,
    ) -> None:
        resp = dict()
        if request.o2_state:
            resp["state"] = request.o2_state

        try:
            resp_type = request.o2_response_type.split()

            code = None
            if "code" in resp_type:
                code = self.code_factory(
                    subject, request.o2_scope, request.o2_client_id, request
                )
                repository.save_authorization_code(
                    code,
                    request.o2_redirect_uri,
                    request.o2_client_id,
                    request,
                )
                resp.update(code.copy())

            token = None
            if "token" in resp_type:
                with_refresh_token = False
                token = self.token_factory(
                    subject, request.o2_scope, with_refresh_token, request
                )
                repository.save_access_token(token, request)
                resp.update(token.copy())

            if "id_token" in resp_type:
                key, alg = repository.get_key_alg_by_client_id(
                    request.o2_client_id, request
                )

                jwt_dict = repository.get_id_token_by_subject(
                    subject, code, token, request
                )
                if request.o2_nonce:
                    jwt_dict["nonce"] = request.o2_nonce
                if code:
                    jwt_dict["c_hash"] = _left_half_hash(code["code"], alg)
                if token:
                    jwt_dict["at_hash"] = _left_half_hash(
                        token["access_token"], alg
                    )
                if code is None and token is None:
                    userinfo = repository.get_userinfo_by_subject(
                        subject, request.o2_scope, request
                    )
                    jwt_dict = jwt_dict.copy()
                    if jwt_dict["sub"] != userinfo["sub"]:
                        raise ServerError(request)
                    jwt_dict.update(userinfo)
                resp["id_token"] = self.jwt_encoder.encode(
                    jwt_dict, key=key, alg=alg
                )
        except BaseException as why:
            raise ServerError(request) from why

        redirect_uri = extend_uri_with_params(
            request.o2_redirect_uri, resp, request.o2_response_mode
        )
        request.response.state = 302
        request.response.location = redirect_uri
        return None

    def supported_grant_types(self) -> List[str]:
        return ["authorization_code"]

    def validate_token_request(
        self,
        request: OpenIDConnectRequestMixin,
        repository: AbstractOpenIDRepository,
        validator: AbstractOpenIDRequestValidator,
    ) -> None:
        super().validate_token_request(request, repository, validator)

        if not request.o2_validate(
            requireds={"code"}, optionals={"redirect_uri"}
        ):
            raise InvalidRequestError(request)

        if not request.o2_validate_redirect_uri(repository, validator):
            raise InvalidGrantError(request)

        if not validator.validate_authorization_code(
            request.o2_code,
            request.o2_redirect_uri,
            request.o2_client_id,
            request,
        ):
            raise InvalidGrantError(request)

    def handle_token_request(
        self,
        request: OpenIDConnectRequestMixin,
        repository: AbstractOpenIDRepository,
    ) -> dict:
        resp = {}
        try:
            subject = repository.get_subject_by_authorization_code(
                request.o2_code, request
            )
            scope = repository.get_scope_by_authorization_code(
                request.o2_code, request
            )
            token = self.token_factory(subject, scope, True, request)
            repository.save_access_token(token, request)
            resp.update(token.copy())

            key, alg = repository.get_key_alg_by_client_id(
                request.o2_client_id, request
            )
            jwt_dict = repository.get_id_token_by_subject(
                subject, {"code": request.o2_code}, token, request
            )
            jwt_dict["at_hash"] = _left_half_hash(token["access_token"], alg)
            resp["id_token"] = self.jwt_encoder.encode(
                jwt_dict, key=key, alg=alg
            )
        except BaseException as why:
            raise ServerError(request) from why

        return resp



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)

from ghoauth.oauth2 import AbstractBaseRepository
from jwt.jwk import AbstractJWKBase

from .request import OpenIDConnectRequestMixin


class AbstractOpenIDRepository(AbstractBaseRepository):

    def get_id_token_by_subject(
        self,
        subject: Any,
        code: Optional[Dict[str, Any]],
        token: Optional[Dict[str, Any]],
        request: OpenIDConnectRequestMixin,
    ) -> Dict[str, Any]:
        raise NotImplementedError("must implement get_id_token_by_subject")

    def get_userinfo_by_subject(
        self, subject: Any, scope: str, request: OpenIDConnectRequestMixin
    ) -> Dict[str, Any]:
        raise NotImplementedError("must implement get_userinfo_by_subject")

    def get_key_alg_by_client_id(
        self, client_id: str, request: OpenIDConnectRequestMixin
    ) -> Tuple[AbstractJWKBase, str]:
        raise NotImplementedError("must implement get_key_alg_by_client_id")



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.oauth2.errors import ServerError
from ghoauth.oauth2.token import bearer_token_factory
from ghoauth.uriutils import extend_uri_with_params

from .abc import AbstractBaseResponseTypeHandler


class ImplicitGrant(AbstractBaseResponseTypeHandler):
    """
    https://tools.ietf.org/html/rfc6749#section-4.2
    """

    __slots__ = ["token_factory"]

    def __init__(self, token_factory=bearer_token_factory):
        self.token_factory = token_factory

    def supported_response_types(self):
        return ["token"]

    def get_default_response_mode(self, request):
        return "fragment"

    def is_response_mode_allowed(self, response_mode, request):
        return response_mode == "fragment"

    def handle_authorization_request(self, request, repository, subject):
        """
        The authorization server MUST NOT issue a refresh token,
        as described in https://tools.ietf.org/html/rfc6749#section-4.2.2
        """
        try:
            token = self.token_factory(
                subject, request.o2_scope, False, request
            )
            repository.save_access_token(token, request)
        except BaseException as why:
            raise ServerError(request) from why

        redirect_uri = extend_uri_with_params(
            request.o2_redirect_uri, token, request.o2_response_mode
        )

        request.response.status = 302
        request.response.location = redirect_uri
        return None



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections.abc import Mapping

from ghoauth.oauth2.errors import (
    InvalidClientError,
    InvalidRequestError,
    UnauthorizedClientError,
    UnsupportedGrantTypeError,
)

from .abc import AbstractBaseEndpoint


class TokenEndpoint(AbstractBaseEndpoint):
    __slots__ = ["repository", "validator", "_handlers"]

    def _get_handler(self, request):
        try:
            return super()._get_handler(request.o2_grant_type)
        except KeyError as why:
            raise UnsupportedGrantTypeError(request) from why

    def register_handler(self, handler):
        # TODO(yosida95): validate the grant_type
        grant_types = handler.supported_grant_types()
        super().register_handler(handler, grant_types)

    def validate_request(self, request):
        """
        A client MAY use the "client_id" request parameter to identify itself
        when sending requests to the token endpoint.  In the
        "authorization_code" "grant_type" request to the token endpoint, an
        unauthenticated client MUST send its "client_id" to prevent itself
        from inadvertently accepting a code intended for a client with a
        different "client_id".
        https://tools.ietf.org/html/rfc6749#section-3.2.1
        """
        if not request.o2_validate(
            requireds={"grant_type"}, optionals={"client_id"}
        ):
            raise InvalidRequestError(request)
        if request.o2_client_id:
            if not self.validator.authenticate_client_as(
                request.o2_client_id, request
            ):
                raise InvalidClientError(request)
        else:
            request.o2_client_id = self.validator.authenticate_client(request)
            if not request.o2_client_id:
                raise InvalidClientError(request)
        if not self.validator.authorize_grant_type(
            request.o2_grant_type, request.o2_client_id, request
        ):
            raise UnauthorizedClientError(request)

        handler = self._get_handler(request)
        handler.validate_token_request(
            request, self.repository, self.validator
        )

    def handle_request(self, request):
        self.validate_request(request)

        handler = self._get_handler(request)
        body = handler.handle_token_request(request, self.repository)

        response = request.response
        if isinstance(body, Mapping):
            response.text = json.dumps(body, ensure_ascii=True, indent=2)
        response.cache_control.no_store = True
        response.pragma = "no-cache"
        return response



# -*- coding: utf-8 -*-
#
# Copyright 2017 Gehirn Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ghoauth.oauth2.errors import (
    FatalOAuth2Error,
    InvalidRequestError,
    InvalidScopeError,
)


class AbstractBaseResponseTypeHandler:

    def supported_response_types(self):
        return []

    def get_default_response_mode(self, request):
        raise NotImplementedError("must implement get_default_response_mode")

    def is_response_mode_allowed(self, response_mode, request):
        raise NotImplementedError("must implement is_response_mode_allowed")

    def validate_authorization_request(self, request, repository, validator):
        if not request.o2_validate(
            requireds={"client_id"}, optionals={"redirect_uri"}
        ):
            raise FatalOAuth2Error()
        if not repository.is_client_available(request.o2_client_id, request):
            raise FatalOAuth2Error()
        if not request.o2_validate_redirect_uri(repository, validator):
            raise FatalOAuth2Error()

        if not request.o2_validate(
            requireds={"response_type"},
            optionals={"scope", "state", "response_mode"},
        ):
            raise InvalidRequestError(request)

        if request.o2_scope:
            if not validator.validate_scope(
                request.o2_scope, request.o2_client_id, request
            ):
                raise InvalidScopeError(request)
        else:
            request.o2_scope = repository.get_default_scope(request)

    def handle_authorization_request(self, request, repository, subject):
        raise NotImplementedError(
            "must implement handle_authorization_request"
        )


class AbstractBaseGrantTypeHandler:

    def supported_grant_types(self):
        return []

    def validate_token_request(self, request, repository, validator):
        return

    def handle_token_request(self, request, repository):
        raise NotImplementedError("must implement handle_token_request")
